' Written by James Deakins, March 2013
' Runs under DOS and regular MMBasic.
' Written using MMBasic V4.3 with a workaround for a bug with MM.DEVICE$. 

' This library reads values from a configuration file.  If the file
' does not exist, it is created and the user is prompted for the values.

' You can allocate a default config file name or leave it blank.
' If blank, the code will try to read the last filename from the
' MM.FNAME$ variable.  If there isn't one there either, the code
' will continue to prompt for a valid file name until it gets one.

' If you use this library, you will have to populate the AllocateValues
' routine with appropriate code to process the parameters and values
' that will be read from the config file.

' Entries in the config file will be in the format
' parameter=value   NOTE: starts 1st column, and no spaces on either 
'                         side of the = sign.
' ' comments        NOTE: the single quote must be on the 1st column.

' Examples
' ' This parameter file holds 3 parameters and their values to support
' ' program xxx.bas.  If you edit this file, all entries start at the 
' ' first column, and there are no spaces on either side of the = sign.
' City=Bangkok
' PostalCode=101010
' Population=000112233

'----------------------------------------------------------------------

False = 0: True = NOT False ' you need these defined!

FileName$ = "Config.cfg"    ' the default, because this is CONFIG.BAS

ReadConfigFile(FileName$) ' as it would appear in your main code
END

'
 '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''' ''''
' Subroutine
' This subroutine reads the user preferences from the configuration
' file.  If the file does not exist, it is created.
' 
' We need two different ways of testing for the file's existance
' because MMBasic only returns valid Error Numbers for files on the
' A: or B: drives.  It does not return a useful error number if 
' the file does not exist in DOS.
' 
' The Config file is the name of the calling program -.BAS +.CFG
'
 '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''' ''''
'
Sub ReadConfigFile(FileName$)
 
Local ConfFile$, Right4$, pFileName$, bValidFileName, LineIn$, ptr, Param$, Value$
LOCAL LineCount, ParamCount, bOK

pFileName$ = FileName$    ' so we don't overwrite the original

IF pFileName$ = "" THEN    ' user has not given a default name to the config file
  pFileName$ = MM.FNAME$   ' so see if the user loaded or saved code as a file name
ENDIF

DO WHILE pFileName$ = ""   ' no default, and no obvious file name, so ask for one
  INPUT "What filename do you want to use for the config file? "; pFileName$
LOOP

' strip off the .bas and add .cfg.  If already .cfg, leave as is. If no extension, add .cfg.

bValidFileName = False
DO WHILE NOT bValidFileName
  bValidFileName = True
  Right4$ = UCASE$(RIGHT$(pFileName$,4))
  IF Right4$ = ".BAS" THEN
    ConfFile$ = LEFT$(pFileName$, LEN(pFileName$) - 4) + ".cfg" ' replace .bas with .cfg
  ELSEIF Right4$ = ".CFG" THEN
    ConfFile$ = pFileName$                                      ' no need to change anything
  ELSEIF INSTR(1,pFileName$,".") = 0 THEN                       ' just a file name, no extension
    ConfFile$ = pFileName$ + ".cfg"
  ELSE
    PRINT "Please give a valid name for your configuration file (with or without an extension)."
    PRINT "You can even give the name of the basic file that has (or needs) configuration data,"
    PRINT "in which case it will use the same file name, but with the .bas extension changed to .cfg."
    PRINT "eg, for a basic program called CONFIG.BAS, supply CONFIG or CONFIG.CFG or CONFIG.BAS"
    PRINT "You can also use multiple configuration file names if you want, like CONFIG1, CONFIG2."    
    PRINT
    pFileName$ = ""
    DO WHILE pFileName$ = ""
      "What filename do you want to use for the config file? "; pFileName$
    LOOP
    bValidFileName$ = False                                     ' force another pass through
  ENDIF 
LOOP

' Open and read the Config file if it exists.  If it doesn't exist, create it then open and read it.  

IF NOT FileExists(ConfFile$) then
   ? "File ";ConfFile$; " does NOT exist"
   ?
   Open ConfFile$ for output as #8
   ' Prompt for the values and create the config file.  Allocate to global variables.
   ? "Please enter your parameter name and the value in the format: "
   ? "parameter=value"
   ? "Continue to enter parameter and value pairs, one pair per line until finished," 
   ? "then just press the Enter key by itself." 
   ? "Comments can also be entered, but they MUST start with a single quote (') "
   ? "as the first character." 
   ? "If you make a mistake, edit the config file with a text editor."
   ?
   LINE INPUT "Next line? - or Enter to finish ", LineIn$
   DO WHILE LineIn$ <> ""
     PRINT #8, LineIn$
     LINE INPUT "Next line? - or Enter to finish ", LineIn$
   LOOP
   ? "Finished entering parameters.  Closing configuration file."  
   Close #8  
ENDIF

' Read in the values - allocate to global variables
bOK        = True
LineCount  = 0
ParamCount = 0
Open ConfFile$ for input as #8
DO WHILE NOT EOF(#8)
  LineCount = LineCount + 1
  LINE INPUT #8, LineIn$
  IF LEFT$(LineIn$,1) <> "'" THEN ' ignore comments
    ptr = INSTR(1, LineIn$, "=")
    Param$ = LEFT$(LineIn$,ptr-1)
    IF ptr = LEN(LineIn$) THEN 
      Value$ = ""                 ' an empty value field
    ELSE
      Value$ = MID$( LineIn$,ptr+1)
    ENDIF
    'the programmer will have to write matching code in function AllocateValues
    if NOT AllocateValues(Param$, Value$) THEN ' returns False if there was a problem
      bOK = False
    ENDIF
    ParamCount = ParamCount + 1
  ENDIF
LOOP

? "Finished reading"; LineCount ; " lines from "; ConfFile$; " and processing"; ParamCount; " parameters."

IF NOT bOK then
    ?
    ? "Errors occurred while processing some parameters."
    ?
    ? "You will have to edit the code in function AllocateValues to process the unknown parameters."
    ? "--------------------------------------------------------------------------------------------"
ENDIF

Close #8  

End Sub ' ReadConfigFile

'
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Function
' Allocates values to global variables depending on the
' values of param$ and value$ passed in
' Returns False if a param cannot be found in the IF statements.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' I know it's obvious, but don't allocate value$ to 
' a local variable :)
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Function AllocateValues(Param$, Value$)

Local UCParam$, bOK

bOK = True

UCParam$ = UCASE$(Param$)

' some examples of parameters and values that might be read in
' and how to allocate them to global values.
' Use an IF statement to process each possible parameter.

IF UCParam$ = "CITY" then
  City$ = Value$                    ' an example for a string value
ELSEIF UCParam$ = "POSTALCODE" then
  PostalCode$ = Value$              ' an example of a numeric value held as a string
ELSEIF UCParam$ = "COUNT" then
  Count = VAL(Value$)               ' an example of converting the string to a numeric value
ELSE
  ? "Unknown parameter and its value "; Param$; "="; Value$
   bOK = False
ENDIF

AllocateValues = bOK

END FUNCTION ' AllocateValues  
'
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Function
' Check whether a file exists.
' Returns a Boolean True or False.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Uses two different methods to check for a file's existence,
' depending on whether the code is run under DOS, or on a 
' Maximite.
' I've submitted a bug report about MM.DEVICE$ returning "" 
' under DOS MMBasic 4.3, and I've used a workaround that
' assumes that "" is only returned if running under DOS.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'
Function fileExists(FileName$)

Local Answer$, bFileExists

if (MM.DEVICE$ = "DOS") OR (MM.DEVICE$ = "") then ' workaround for bug - see comment above.
  open "exist.bat" for output as #8
  Print #8, "if exist " + chr$(34) + FileName$ + chr$(34) + " ( "
  Print #8, "   echo YES> exist.tmp"
  Print #8, ") Else ("
  Print #8, "   echo NO> exist.tmp"
  Print #8, ")"
  Print #8, "Exit" 
  Close #8

  ' Run batch that redirects DIR results to exist.tmp file.
  ' Redirect all messages from the batch file to nul (ie, not displayed)
  SYSTEM "exist.bat > nul"

  ' Read the single line from exist.tmp for the answer as to whether the file exists.
  Open "exist.tmp" for input as #8
  Line Input #8, Answer$
  bFileExists = (Answer$ <> "YES")
  Close #8
  KILL "exist.bat"  ' and clean up 
  KILL "exist.tmp"  
else      ' running on a Maximite or similar hardware   
' Turn off Abort on Error so we can capture the Error Number
  Option Error Continue
  Open FileName$ for Input as #8 
   
  If MM.ERRNO = 6 then ' can't find file
    bFileExists = False
  elseif MM.ERRNO = 0 then
    bFileExists = True
    CLOSE #8
  else
    ? "ERROR: the Existence Check for file "; ConfFile$; " returned MM.ERRNO "; MM.ERRNO
    ? "The description of MM.ERRNO"; MM.ERRNO; " is: "; ErrNoDesc$(MM.ERRNO)
    ' Close #8
    Option Error Abort
    END
  Endif
  Option Error Abort
Endif

fileExists = bFileExists
 
End Function ' fileExists

'
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Function
' Return the matching Description for an MMBasic 
' Error Number.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' The descriptions  were taken from the V4.3 manual.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'       
Function ErrNoDesc$(ErrNo)
  
Local Desc$
 
if ErrNo = 0 then
  Desc$ = "No error"
elseif ErrNo = 1 then
  Desc$ = "No SD card found"
elseif ErrNo = 2 then
  Desc$ = "SD card is write protected"
elseif ErrNo = 3 then
  Desc$ = "Not enough space"
elseif ErrNo = 4 then
  Desc$ = "All root directory entries are taken"
elseif ErrNo = 5 then
  Desc$ = "Invalid filename"
elseif ErrNo = 6 then
  Desc$ = "Cannot find file"
elseif ErrNo = 7 then
  Desc$ = "Cannot find directory"
elseif ErrNo = 8 then
  Desc$ = "File is read only"
elseif ErrNo = 9 then
  Desc$ = "Cannot open file"
elseif ErrNo = 10 then
  Desc$ = "Error reading from file"
elseif ErrNo = 11 then
  Desc$ = "Error writing to file"
elseif ErrNo = 12 then
  Desc$ = "Not a file"
elseif ErrNo = 13 then
  Desc$ = "Not a directory"
elseif ErrNo = 14 then
  Desc$ = "Directory not empty"
elseif ErrNo = 15 then
  Desc$ = "Hardware error accessing the storage media"
elseif ErrNo = 16 then
  Desc$ = "Flash memory write failure"
else
  Desc$ = "******** " + ErrNo + " is an invalid Error Number iaw V4.3 manual **********"
endif        
 
ErrNoDesc$ = Desc$

End Function ' ErrNoDesc$ 
